@@ -1,8 +1,8 @@
 /*
-** oidentd_forward.c - oidentd request forwarding.
+** forward.c - oidentd request forwarding.
 ** Copyright (c) 1998-2006 Ryan McCabe <ryan@numb.org>
 ** Copyright (c) 2018      Jan Steffens <jan.steffens@gmail.com>
-** Copyright (c) 2018-2019 Janik Rabe  <oidentd@janikrabe.com>
+** Copyright (c) 2018-2019 Janik Rabe  <info@janikrabe.com>
 **
 ** This program is free software; you can redistribute it and/or modify
 ** it under the terms of the GNU General Public License, version 2,
@@ -86,7 +86,7 @@ int forward_request(const struct sockaddr_storage *host,
 
 	alarm(5);
 
-	if (connect(fsock, (struct sockaddr *) &addr, sin_len(&addr)) != 0) {
+	if (connect(fsock, (struct sockaddr *) &addr, (socklen_t) sin_len(&addr)) != 0) {
 		get_ip(&addr, ipbuf, sizeof(ipbuf));
 		debug("connect to %s:%d: %s",
 			ipbuf, ntohs(sin_port(&addr)), strerror(errno));
@@ -99,7 +99,7 @@ int forward_request(const struct sockaddr_storage *host,
 	}
 
 	if (!sock_read(fsock, buf, sizeof(buf))) {
-		debug("read(%d): %s\n", fsock, strerror(errno));
+		debug("read(%d): %s", fsock, strerror(errno));
 		goto out_fail;
 	}
 
@@ -135,7 +135,7 @@ out_fail:
 */
 
 static void fwd_alarm(int sig) {
-	o_log(NORMAL, "Forward timed out");
+	o_log(LOG_INFO, "Forward timed out");
 	close(fsock);
 	siglongjmp(timebuf, sig);
 }
