@@ -1,7 +1,7 @@
 /*
 ** oidentd.c - oidentd ident (RFC 1413) implementation.
 ** Copyright (c) 1998-2006 Ryan McCabe <ryan@numb.org>
-** Copyright (c) 2018-2019 Janik Rabe  <oidentd@janikrabe.com>
+** Copyright (c) 2018-2019 Janik Rabe  <info@janikrabe.com>
 **
 ** This program is free software; you can redistribute it and/or modify
 ** it under the terms of the GNU General Public License, version 2,
@@ -46,11 +46,6 @@
 #include "options.h"
 #include "masq.h"
 
-#if HAVE_LIBUDB
-#	warning "libudb support is deprecated"
-#	include <udb.h>
-#endif
-
 #ifndef FUZZING_BUILD_MODE_UNSAFE_FOR_PRODUCTION
 static void sig_segv(int unused __notused) __noreturn;
 static void sig_child(int sig);
@@ -72,6 +67,7 @@ gid_t target_gid;
 
 char *ret_os;
 char *failuser;
+char *replyall;
 char *config_file;
 
 in_port_t listen_port;
@@ -85,12 +81,12 @@ int main(int argc, char **argv) {
 
 	openlog(PACKAGE_NAME, LOG_PID | LOG_CONS | LOG_NDELAY, LOG_DAEMON);
 
-	if (read_config(config_file) != 0) {
+	if (!replyall && read_config(config_file) != 0) {
 		o_log(LOG_CRIT, "Fatal: Error reading configuration file");
 		exit(EXIT_FAILURE);
 	}
 
-	if (core_init() != 0) {
+	if (!replyall && core_init() != 0) {
 		if (opt_enabled(DEBUG_MSGS)) {
 			o_log(LOG_CRIT, "Fatal: Error initializing core");
 		} else {
@@ -113,7 +109,7 @@ int main(int argc, char **argv) {
 		exit(EXIT_FAILURE);
 	}
 
-	if (k_open() != 0) {
+	if (!replyall && k_open() != 0) {
 		o_log(LOG_CRIT, "Fatal: Unable to initialize kernel module: %s", strerror(errno));
 		exit(EXIT_FAILURE);
 	}
@@ -123,15 +119,6 @@ int main(int argc, char **argv) {
 		exit(EXIT_FAILURE);
 	}
 
-#if HAVE_LIBUDB
-	if (opt_enabled(USEUDB)) {
-		if (udb_init(UDB_ENV_BASE_KEY) == 0) {
-			o_log(LOG_CRIT, "Fatal: Can't open UDB shared memory tables");
-			exit(EXIT_FAILURE);
-		}
-	}
-#endif
-
 #ifndef FUZZING_BUILD_MODE_UNSAFE_FOR_PRODUCTION
 	signal(SIGALRM, sig_alarm);
 	signal(SIGCHLD, sig_child);
@@ -291,22 +278,22 @@ static int service_request(int insock, int outsock) {
 		return 0;
 	}
 
+	if (replyall) {
+		sockprintf(outsock, "%d,%d:USERID:%s:%s\r\n",
+			lport_temp, fport_temp, ret_os, replyall);
+
+		o_log(LOG_INFO, "[%s] Static reply: %d , %d : (returned %s)",
+			host_buf, lport_temp, fport_temp, replyall);
+
+		return 0;
+	}
+
 	lport = (in_port_t) lport_temp;
 	fport = (in_port_t) fport_temp;
 
 	/* User ID is unknown. */
 	con_uid = MISSING_UID;
 
-#if HAVE_LIBUDB
-	if (opt_enabled(USEUDB)) {
-		struct udb_lookup_res udb_res = get_udb_user(
-				lport, fport, &laddr, &faddr, insock);
-		if (udb_res.status == 2)
-			return 0;
-		con_uid = udb_res.uid;
-	}
-#endif
-
 	if (con_uid == MISSING_UID && laddr.ss_family == AF_INET)
 		con_uid = get_user4(htons(lport), htons(fport), &laddr, &faddr);
 
@@ -452,6 +439,13 @@ static void sig_alarm(int unused __notused) {
 */
 
 static void sig_hup(int unused __notused) {
+	if (replyall) {
+		/*
+		 * Configuration file was never loaded; don't try to reload it.
+		 */
+		return;
+	}
+
 	user_db_destroy();
 
 	if (read_config(CONFFILE) != 0) {
