@@ -1,7 +1,7 @@
 /*
 ** linux.c - Linux user lookup facility.
 ** Copyright (c) 1998-2006 Ryan McCabe <ryan@numb.org>
-** Copyright (c) 2018      Janik Rabe  <oidentd@janikrabe.com>
+** Copyright (c) 2018-2019 Janik Rabe  <info@janikrabe.com>
 **
 ** This program is free software; you can redistribute it and/or modify
 ** it under the terms of the GNU General Public License, version 2,
@@ -43,38 +43,23 @@
 #include "options.h"
 #include "netlink.h"
 
-#ifdef HAVE_LIBCAP_NG
-#	include <cap-ng.h>
-#else
+#if !MASQ_SUPPORT
 #	undef LIBNFCT_SUPPORT
 #endif
 
-#ifndef MASQ_SUPPORT
-#	undef LIBNFCT_SUPPORT
-#endif
-
-#ifdef LIBNFCT_SUPPORT
+#if LIBNFCT_SUPPORT
 #	include <libnetfilter_conntrack/libnetfilter_conntrack.h>
 #endif
 
-#ifdef HAVE_LIBUDB
-#	include <udb.h>
-#endif
-
 #define CFILE		"/proc/net/tcp"
 #define CFILE6		"/proc/net/tcp6"
-#define MASQFILE	"/proc/net/ip_masquerade"
-#define IPCONNTRACK	"/proc/net/ip_conntrack"
 #define NFCONNTRACK	"/proc/net/nf_conntrack"
 
 static int netlink_sock;
 extern struct sockaddr_storage proxy;
 extern char *ret_os;
 
-extern uid_t uid;
-extern gid_t gid;
-
-#ifdef LIBNFCT_SUPPORT
+#if LIBNFCT_SUPPORT
 struct ct_masq_query {
 	int sock;
 	in_port_t lport;
@@ -85,17 +70,17 @@ struct ct_masq_query {
 };
 #endif
 
-#ifdef MASQ_SUPPORT
+#if MASQ_SUPPORT
 static int masq_ct_line(char *line,
 			int sock,
+			int ct_type,
 			in_port_t lport,
 			in_port_t fport,
 			struct sockaddr_storage *laddr,
 			struct sockaddr_storage *faddr);
 #endif
 
-#ifdef LIBNFCT_SUPPORT
-bool drop_privs_libnfct(uid_t uid, gid_t gid);
+#if LIBNFCT_SUPPORT
 static bool dispatch_libnfct_query(struct ct_masq_query *queryp);
 static int callback_nfct(enum nf_conntrack_msg_type type,
 			struct nf_conntrack *ct,
@@ -107,52 +92,16 @@ static uid_t lookup_tcp_diag(	struct sockaddr_storage *src_addr,
 							in_port_t src_port,
 							in_port_t dst_port);
 
-#ifdef MASQ_SUPPORT
+#if MASQ_SUPPORT
 enum {
 	CT_UNKNOWN,
-	CT_MASQFILE,
-	CT_IPCONNTRACK,
 	CT_NFCONNTRACK,
-#	ifdef LIBNFCT_SUPPORT
-	CT_LIBNFCT,
-#	endif
 };
 FILE *masq_fp;
 static int conntrack = CT_UNKNOWN;
 #endif
 
-#ifdef LIBNFCT_SUPPORT
-bool drop_privs_libnfct(uid_t uid, gid_t gid) {
-	if (conntrack != CT_LIBNFCT)
-		return true;
-
-	/* drop privileges, keeping only CAP_NET_ADMIN for libnfct queries */
-
-	int ret;
-	capng_clear(CAPNG_SELECT_BOTH);
-	ret = capng_update(CAPNG_ADD, CAPNG_EFFECTIVE | CAPNG_PERMITTED,
-			CAP_NET_ADMIN);
-	if (ret) {
-		debug("capng_update: error %d", ret);
-		return false;
-	}
-
-	ret = capng_change_id(
-			opt_enabled(CHANGE_UID) ? uid : MISSING_UID,
-			opt_enabled(CHANGE_GID) ? gid : MISSING_GID,
-			CAPNG_CLEAR_BOUNDING | CAPNG_DROP_SUPP_GRP);
-	if (ret) {
-		debug("capng_change_id: error %d", ret);
-		return false;
-	}
-
-	/* don't try to drop privileges again */
-	disable_opt(CHANGE_UID);
-	disable_opt(CHANGE_GID);
-
-	return true;
-}
-
+#if LIBNFCT_SUPPORT
 static bool dispatch_libnfct_query(struct ct_masq_query *queryp) {
 	struct nfct_handle *nfcthp = nfct_open(CONNTRACK, 0);
 
@@ -177,22 +126,25 @@ static bool dispatch_libnfct_query(struct ct_masq_query *queryp) {
 		return false;
 	}
 
-	return (!queryp->status);
+	return !queryp->status;
 }
 
 /*
 ** Callback for libnetfilter_conntrack queries
 */
 
-static int callback_nfct(enum nf_conntrack_msg_type type,
+static int callback_nfct(enum nf_conntrack_msg_type type __notused,
 			struct nf_conntrack *ct,
 			void *data) {
 	char buf[1024];
+	struct ct_masq_query *query;
+	int ret;
+
 	nfct_snprintf(buf, sizeof(buf), ct, NFCT_T_UNKNOWN,
 			NFCT_O_DEFAULT, NFCT_OF_SHOW_LAYER3);
 
-	struct ct_masq_query *query = (struct ct_masq_query *) data;
-	int ret = masq_ct_line(buf, query->sock,
+	query = (struct ct_masq_query *) data;
+	ret = masq_ct_line(buf, query->sock, CT_NFCONNTRACK,
 			query->lport, query->fport,
 			query->laddr, query->faddr);
 
@@ -202,69 +154,46 @@ static int callback_nfct(enum nf_conntrack_msg_type type,
 	query->status = ret;
 	return NFCT_CB_STOP;
 }
-
 #endif
 
 /*
 ** System-dependent initialization; called only once.
 ** Called before privileges are dropped.
-** Returns false on failure.
+** Returns non-zero on failure.
 */
 
-bool core_init(void) {
-#ifdef MASQ_SUPPORT
+int core_init(void) {
+#if MASQ_SUPPORT
 	if (!opt_enabled(MASQ)) {
 		masq_fp = NULL;
-		return true;
+		return 0;
 	}
 
-	masq_fp = fopen(MASQFILE, "r");
-	if (masq_fp == NULL) {
+	masq_fp = fopen(NFCONNTRACK, "r");
+	if (!masq_fp) {
 		if (errno != ENOENT) {
-			o_log(LOG_CRIT, "fopen: %s: %s", MASQFILE, strerror(errno));
-			return false;
+			o_log(LOG_CRIT, "fopen: %s: %s", NFCONNTRACK, strerror(errno));
+			return -1;
 		}
 
-		masq_fp = fopen(NFCONNTRACK, "r");
-		if (masq_fp == NULL) {
-			if (errno != ENOENT) {
-				o_log(LOG_CRIT, "fopen: %s: %s", NFCONNTRACK, strerror(errno));
-				return false;
-			}
-
-			masq_fp = fopen(IPCONNTRACK, "r");
-			if (masq_fp == NULL) {
-				if (errno != ENOENT) {
-					o_log(LOG_CRIT, "fopen: %s: %s", IPCONNTRACK, strerror(errno));
-					return false;
-				}
-
-#	ifdef LIBNFCT_SUPPORT
-				conntrack = CT_LIBNFCT;
-				return true;
+#	if LIBNFCT_SUPPORT
+		return 0;
 #	else
-				o_log(LOG_CRIT, "NAT/IP masquerading support is unavailable");
-				disable_opt(MASQ);
+		o_log(LOG_CRIT, "NAT/IP masquerading support is unavailable "
+						"because " PACKAGE_NAME " was compiled without "
+						"libnetfilter_conntrack support and no "
+						"connection tracking file was found.");
+		disable_opt(MASQ);
 #	endif
-			} else {
-				conntrack = CT_IPCONNTRACK;
-			}
-		} else {
-			conntrack = CT_NFCONNTRACK;
-		}
-	} else if (opt_enabled(PROXY) || opt_enabled(FORWARD)) {
-		o_log(LOG_CRIT, "Only local NAT is supported on your system; "
-		                "please consider upgrading your kernel");
-		return false;
 	} else {
-		conntrack = CT_MASQFILE;
+		conntrack = CT_NFCONNTRACK;
 	}
 #endif
-	return true;
+	return 0;
 }
 
 
-#ifdef WANT_IPV6
+#if WANT_IPV6
 
 /*
 ** Returns the UID of the owner of an IPv6 connection,
@@ -283,20 +212,20 @@ uid_t get_user6(	in_port_t lport,
 		uid_t uid = lookup_tcp_diag(laddr, faddr, lport, fport);
 
 		if (uid != MISSING_UID)
-			return (uid);
+			return uid;
 	}
 
 	lport = ntohs(lport);
 	fport = ntohs(fport);
 
 	fp = fopen(CFILE6, "r");
-	if (fp == NULL) {
+	if (!fp) {
 		debug("fopen: %s: %s", CFILE6, strerror(errno));
 		return MISSING_UID;
 	}
 
 	/* Eat the header line. */
-	if (NULL == fgets(buf, sizeof(buf), fp)) {
+	if (!fgets(buf, sizeof(buf), fp)) {
 		debug("fgets: %s: Could not read header", CFILE6);
 		fclose(fp);
 		return MISSING_UID;
@@ -309,12 +238,12 @@ uid_t get_user6(	in_port_t lport,
 		u_int32_t portf_temp;
 		in_port_t portl;
 		in_port_t portf;
-		uid_t uid;
+		unsigned long uid;
+		unsigned long inode;
 		int ret;
-		u_long inode;
 
 		ret = sscanf(buf,
-			"%*d: %8x%8x%8x%8x:%x %8x%8x%8x%8x:%x %*x %*X:%*X %*x:%*X %*x %u %*d %lu",
+			"%*d: %8x%8x%8x%8x:%x %8x%8x%8x%8x:%x %*x %*X:%*X %*x:%*X %*x %lu %*d %lu",
 			&local6.s6_addr32[0], &local6.s6_addr32[1], &local6.s6_addr32[2],
 			&local6.s6_addr32[3], &portl_temp,
 			&remote6.s6_addr32[0], &remote6.s6_addr32[1], &remote6.s6_addr32[2],
@@ -337,7 +266,7 @@ uid_t get_user6(	in_port_t lport,
 			if (inode == 0 && uid == 0)
 				return MISSING_UID;
 
-			return (uid);
+			return (uid_t) uid;
 		}
 	}
 
@@ -357,18 +286,18 @@ uid_t get_user4(	in_port_t lport,
 				struct sockaddr_storage *laddr,
 				struct sockaddr_storage *faddr)
 {
-	uid_t uid;
+	unsigned long uid;
+	unsigned long inode;
 	FILE *fp;
 	char buf[1024];
-	u_int32_t inode;
 	in_addr_t laddr4;
 	in_addr_t faddr4;
 
 	if (netlink_sock != -1) {
-		uid = lookup_tcp_diag(laddr, faddr, lport, fport);
+		uid_t nluid = lookup_tcp_diag(laddr, faddr, lport, fport);
 
-		if (uid != MISSING_UID)
-			return (uid);
+		if (nluid != MISSING_UID)
+			return nluid;
 	}
 
 	laddr4 = SIN4(laddr)->sin_addr.s_addr;
@@ -378,13 +307,13 @@ uid_t get_user4(	in_port_t lport,
 	fport = ntohs(fport);
 
 	fp = fopen(CFILE, "r");
-	if (fp == NULL) {
+	if (!fp) {
 		debug("fopen: %s: %s", CFILE, strerror(errno));
 		return MISSING_UID;
 	}
 
 	/* Eat the header line. */
-	if (NULL == fgets(buf, sizeof(buf), fp)) {
+	if (!fgets(buf, sizeof(buf), fp)) {
 		debug("fgets: %s: Could not read header", CFILE);
 		fclose(fp);
 		return MISSING_UID;
@@ -404,7 +333,7 @@ uid_t get_user4(	in_port_t lport,
 		in_addr_t remote;
 
 		ret = sscanf(buf,
-			"%*d: %x:%x %x:%x %*x %*x:%*x %*x:%*x %*x %u %*d %u",
+			"%*d: %x:%x %x:%x %*x %*x:%*x %*x:%*x %*x %lu %*d %lu",
 			&local, &portl_temp, &remote, &portf_temp, &uid, &inode);
 
 		if (ret != 6)
@@ -447,74 +376,65 @@ out_success:
 	if (inode == 0 && uid == 0)
 		return MISSING_UID;
 
-	return (uid);
+	return (uid_t) uid;
 }
 
-#ifdef MASQ_SUPPORT
+#if MASQ_SUPPORT
 
 /*
 ** Handle a request to a host that's IP masquerading through us.
-** Returns true on success, false on failure.
+** Returns non-zero on failure.
 */
 
-bool masq(	int sock,
+int masq(	int sock,
 			in_port_t lport,
 			in_port_t fport,
 			struct sockaddr_storage *laddr,
 			struct sockaddr_storage *faddr)
 {
 	char buf[1024];
-
-	/*
-	** There's no masq support for IPv6 yet.
-	*/
-
-	if (faddr->ss_family != AF_INET)
-		return false;
+#if LIBNFCT_SUPPORT
+	struct ct_masq_query query;
+#endif
 
 	lport = ntohs(lport);
 	fport = ntohs(fport);
 
-#ifdef LIBNFCT_SUPPORT
-	if (conntrack == CT_LIBNFCT) {
-		struct ct_masq_query query = { sock,
-				lport, fport,
-				laddr, faddr, 1 };
-		return dispatch_libnfct_query(&query);
-	}
+#if LIBNFCT_SUPPORT
+	query = (struct ct_masq_query) { sock, lport, fport, laddr, faddr, 1 };
+
+	if (dispatch_libnfct_query(&query))
+		return 0;
 #endif
 
-	/* rewind fp to read new contents */
-	rewind(masq_fp);
+	if (masq_fp) {
+		/* rewind fp to read new contents */
+		rewind(masq_fp);
 
-	if (conntrack == CT_MASQFILE) {
-		/* eat the header line */
-		if (NULL == fgets(buf, sizeof(buf), masq_fp)) {
-			debug("fgets: conntrack file: Could not read header");
-			return false;
+		while (fgets(buf, sizeof(buf), masq_fp)) {
+			int ret = masq_ct_line(buf, sock, conntrack,
+				lport, fport, laddr, faddr);
+			if (ret == 1)
+				continue;
+			return ret;
 		}
-	}
+	} else if (conntrack != CT_UNKNOWN)
+		debug("Connection tracking file is in use but not open");
 
-	while (fgets(buf, sizeof(buf), masq_fp)) {
-		int ret = masq_ct_line(buf, sock, lport, fport, laddr, faddr);
-		if (ret == 1)
-			continue;
-		return !ret;
-	}
-
-	return false;
+	return -1;
 }
 
 /*
 ** Process a connection tracking file entry.
 ** The lport and fport arguments are in host byte order.
-** Returns -1 if an error occurs.
-** Returns  0 if the entry matches and the request has been handled.
-** Returns  1 if the entry does not match the query.
+** Returns -1 if an error occurred.
+** Returns  0 if the entry matched and the request has been handled.
+** Returns  1 if the entry did not match the query.
 **/
 
 static int masq_ct_line(char *line,
 			int sock,
+			int ct_type,
 			in_port_t lport,
 			in_port_t fport,
 			struct sockaddr_storage *laddr,
@@ -527,118 +447,96 @@ static int masq_ct_line(char *line,
 	in_port_t masq_lport;
 	in_port_t masq_fport;
 	char user[MAX_ULEN];
-	in_addr_t localm;
-	in_addr_t remotem;
-	in_addr_t localn;
-	in_addr_t remoten;
-	struct sockaddr_storage ss;
+	struct sockaddr_storage localm_ss;
+	struct sockaddr_storage remotem_ss;
+	struct sockaddr_storage localn_ss;
+	struct sockaddr_storage remoten_ss;
 	int ret;
 
-	if (conntrack == CT_MASQFILE) {
-		u_int32_t mport_temp;
-		u_int32_t nport_temp;
-		u_int32_t masq_lport_temp;
-		u_int32_t masq_fport_temp;
-
-		ret = sscanf(line, "%15s %X:%X %X:%X %X %X %*d %*d %*u",
-				proto, &localm, &masq_lport_temp,
-				&remotem, &masq_fport_temp, &mport_temp, &nport_temp);
-
-		if (ret != 7)
-			return 1;
-
-		mport = (in_port_t) mport_temp;
-		nport = (in_port_t) nport_temp;
-		masq_lport = (in_port_t) masq_lport_temp;
-		masq_fport = (in_port_t) masq_fport_temp;
-
-		/* Assume local NAT. */
-		remoten = localm;
-		localn = remotem;
-	} else if (conntrack == CT_IPCONNTRACK) {
-		int l1, l2, l3, l4, r1, r2, r3, r4;
-		int nl1, nl2, nl3, nl4, nr1, nr2, nr3, nr4;
+	if (ct_type == CT_NFCONNTRACK) {
+		char ml[MAX_IPLEN];
+		char mr[MAX_IPLEN];
+		char nl[MAX_IPLEN];
+		char nr[MAX_IPLEN];
+		in_addr_t localm4;
+		in_addr_t remotem4;
+		in_addr_t localn4;
+		in_addr_t remoten4;
+		struct in6_addr localm6;
+		struct in6_addr remotem6;
+		struct in6_addr localn6;
+		struct in6_addr remoten6;
 		u_int32_t nport_temp;
 		u_int32_t mport_temp;
 		u_int32_t masq_lport_temp;
 		u_int32_t masq_fport_temp;
 
 		ret = sscanf(line,
-			"%15s %*d %*d ESTABLISHED src=%d.%d.%d.%d dst=%d.%d.%d.%d sport=%d dport=%d src=%d.%d.%d.%d dst=%d.%d.%d.%d sport=%d dport=%d",
-			proto, &l1, &l2, &l3, &l4, &r1, &r2, &r3, &r4,
-			&masq_lport_temp, &masq_fport_temp,
-			&nl1, &nl2, &nl3, &nl4, &nr1, &nr2, &nr3, &nr4,
-			&nport_temp, &mport_temp);
-
-		if (ret != 21) {
-			ret = sscanf(line,
-				"%15s %*d %*d ESTABLISHED src=%d.%d.%d.%d dst=%d.%d.%d.%d sport=%d dport=%d packets=%*d bytes=%*d src=%d.%d.%d.%d dst=%d.%d.%d.%d sport=%d dport=%d",
-			proto, &l1, &l2, &l3, &l4, &r1, &r2, &r3, &r4,
-			&masq_lport_temp, &masq_fport_temp,
-			&nl1, &nl2, &nl3, &nl4, &nr1, &nr2, &nr3, &nr4,
-			&nport_temp, &mport_temp);
-		}
-
-		if (ret != 21)
-			return 1;
-
-		masq_lport = (in_port_t) masq_lport_temp;
-		masq_fport = (in_port_t) masq_fport_temp;
-
-		nport = (in_port_t) nport_temp;
-		mport = (in_port_t) mport_temp;
-
-		localm = l1 << 24 | l2 << 16 | l3 << 8 | l4;
-		remotem = r1 << 24 | r2 << 16 | r3 << 8 | r4;
-
-		localn = nl1 << 24 | nl2 << 16 | nl3 << 8 | nl4;
-		remoten = nr1 << 24 | nr2 << 16 | nr3 << 8 | nr4;
-#ifdef LIBNFCT_SUPPORT
-	} else if (conntrack == CT_NFCONNTRACK || conntrack == CT_LIBNFCT) {
-#else
-	} else if (conntrack == CT_NFCONNTRACK) {
-#endif
-		int l1, l2, l3, l4, r1, r2, r3, r4;
-		int nl1, nl2, nl3, nl4, nr1, nr2, nr3, nr4;
-		u_int32_t nport_temp;
-		u_int32_t mport_temp;
-		u_int32_t masq_lport_temp;
-		u_int32_t masq_fport_temp;
-
-		ret = sscanf(line,
-			"%15s %*d %15s %*d %*d ESTABLISHED src=%d.%d.%d.%d dst=%d.%d.%d.%d sport=%d dport=%d packets=%*d bytes=%*d src=%d.%d.%d.%d dst=%d.%d.%d.%d sport=%d dport=%d",
-			family, proto, &l1, &l2, &l3, &l4, &r1, &r2, &r3, &r4,
-			&masq_lport_temp, &masq_fport_temp,
-			&nl1, &nl2, &nl3, &nl4, &nr1, &nr2, &nr3, &nr4,
-			&nport_temp, &mport_temp);
+			"%15s %*d %15s %*d %*d ESTABLISHED"
+			" src=" IP_SCAN_SPEC " dst=" IP_SCAN_SPEC " sport=%d dport=%d"
+			" packets=%*d bytes=%*d"
+			" src=" IP_SCAN_SPEC " dst=" IP_SCAN_SPEC " sport=%d dport=%d",
+			family, proto,
+			ml, mr, &masq_lport_temp, &masq_fport_temp,
+			nl, nr, &nport_temp, &mport_temp);
 
 		/* Added to handle /proc/sys/net/netfilter/nf_conntrack_acct = 0 */
-		if (ret != 22) {
+		if (ret != 10) {
 			ret = sscanf(line,
-				"%15s %*d %15s %*d %*d ESTABLISHED src=%d.%d.%d.%d dst=%d.%d.%d.%d sport=%d dport=%d src=%d.%d.%d.%d dst=%d.%d.%d.%d sport=%d dport=%d",
-				family, proto, &l1, &l2, &l3, &l4, &r1, &r2, &r3, &r4,
-				&masq_lport_temp, &masq_fport_temp,
-				&nl1, &nl2, &nl3, &nl4, &nr1, &nr2, &nr3, &nr4,
-				&nport_temp, &mport_temp);
+				"%15s %*d %15s %*d %*d ESTABLISHED"
+				" src=" IP_SCAN_SPEC " dst=" IP_SCAN_SPEC " sport=%d dport=%d"
+				" src=" IP_SCAN_SPEC " dst=" IP_SCAN_SPEC " sport=%d dport=%d",
+				family, proto,
+				ml, mr, &masq_lport_temp, &masq_fport_temp,
+				nl, nr, &nport_temp, &mport_temp);
 		}
 
-		if (ret != 22)
+		if (ret != 10)
 			return 1;
 
-		if (strcasecmp(family, "ipv4"))
-			return 1;
+		switch (faddr->ss_family) {
+		case AF_INET:
+			if (strcasecmp(family, "ipv4"))
+				return 1;
+
+			if (inet_pton(AF_INET, ml, &localm4)  < 0 ||
+			    inet_pton(AF_INET, mr, &remotem4) < 0 ||
+			    inet_pton(AF_INET, nl, &localn4)  < 0 ||
+			    inet_pton(AF_INET, nr, &remoten4) < 0)
+				return 1;
+
+			sin_setv4(localm4, &localm_ss);
+			sin_setv4(remotem4, &remotem_ss);
+			sin_setv4(localn4, &localn_ss);
+			sin_setv4(remoten4, &remoten_ss);
+
+			break;
+		case AF_INET6:
+			if (strcasecmp(family, "ipv6"))
+				return 1;
+
+			if (inet_pton(AF_INET6, ml, &localm6)  < 0 ||
+			    inet_pton(AF_INET6, mr, &remotem6) < 0 ||
+			    inet_pton(AF_INET6, nl, &localn6)  < 0 ||
+			    inet_pton(AF_INET6, nr, &remoten6) < 0)
+				return 1;
+
+			sin_setv6(&localm6, &localm_ss);
+			sin_setv6(&remotem6, &remotem_ss);
+			sin_setv6(&localn6, &localn_ss);
+			sin_setv6(&remoten6, &remoten_ss);
+
+			break;
+		default:
+			debug("masq_ct_line: bad address family %d", faddr->ss_family);
+			return -1;
+		}
 
 		masq_lport = (in_port_t) masq_lport_temp;
 		masq_fport = (in_port_t) masq_fport_temp;
 
 		nport = (in_port_t) nport_temp;
 		mport = (in_port_t) mport_temp;
-
-		localm = l1 << 24 | l2 << 16 | l3 << 8 | l4;
-		remotem = r1 << 24 | r2 << 16 | r3 << 8 | r4;
-
-		localn = nl1 << 24 | nl2 << 16 | nl3 << 8 | nl4;
-		remoten = nr1 << 24 | nr2 << 16 | nr3 << 8 | nr4;
 	} else
 		return -1;
 
@@ -652,38 +550,38 @@ static int masq_ct_line(char *line,
 		return 1;
 
 	/* Local NAT, don't forward or do masquerade entry lookup. */
-	if (localm == remoten) {
+	if (sin_equal(&localm_ss, &remoten_ss)) {
 		uid_t con_uid = MISSING_UID;
 		struct passwd *pw;
 		char suser[MAX_ULEN];
 		char ipbuf[MAX_IPLEN];
 
-		sin_setv4(htonl(remotem), &ss);
 		get_ip(faddr, ipbuf, sizeof(ipbuf));
 
 		if (con_uid == MISSING_UID && faddr->ss_family == AF_INET)
-			con_uid = get_user4(htons(masq_lport), htons(masq_fport), laddr, &ss);
+			con_uid = get_user4(htons(masq_lport), htons(masq_fport), laddr, &remotem_ss);
 
-		/* Add call to get_user6 when IPv6 NAT is supported. */
+		if (con_uid == MISSING_UID && faddr->ss_family == AF_INET6)
+			con_uid = get_user6(htons(masq_lport), htons(masq_fport), laddr, &remotem_ss);
 
 		if (con_uid == MISSING_UID)
 			return -1;
 
 		pw = getpwuid(con_uid);
-		if (pw == NULL) {
+		if (!pw) {
 			sockprintf(sock, "%d,%d:ERROR:%s\r\n",
 				lport, fport, ERROR("NO-USER"));
 
-			debug("getpwuid(%u): %s", con_uid, strerror(errno));
+			debug("getpwuid(%lu): %s", (unsigned long) con_uid, strerror(errno));
 			return 0;
 		}
 
-		ret = get_ident(pw, masq_lport, masq_fport, laddr, &ss, suser, sizeof(suser));
+		ret = get_ident(pw, masq_lport, masq_fport, laddr, &remotem_ss, suser, sizeof(suser));
 		if (ret == -1) {
 			sockprintf(sock, "%d,%d:ERROR:%s\r\n",
 				lport, fport, ERROR("HIDDEN-USER"));
 
-			o_log(NORMAL, "[%s] %d (%d) , %d (%d) : HIDDEN-USER (%s)",
+			o_log(LOG_INFO, "[%s] %d (%d) , %d (%d) : HIDDEN-USER (%s)",
 				ipbuf, lport, masq_lport, fport, masq_fport, pw->pw_name);
 
 			return 0;
@@ -692,34 +590,32 @@ static int masq_ct_line(char *line,
 		sockprintf(sock, "%d,%d:USERID:%s:%s\r\n",
 			lport, fport, ret_os, suser);
 
-		o_log(NORMAL, "[%s] Successful lookup: %d (%d) , %d (%d) : %s (%s)",
+		o_log(LOG_INFO, "[%s] Successful lookup: %d (%d) , %d (%d) : %s (%s)",
 			ipbuf, lport, masq_lport, fport, masq_fport, pw->pw_name, suser);
 
 		return 0;
 	}
 
-	if (localn != ntohl(SIN4(faddr)->sin_addr.s_addr)) {
+	if (!sin_equal(&localn_ss, faddr)) {
 		if (!opt_enabled(PROXY))
 			return 1;
 
-		if (SIN4(faddr)->sin_addr.s_addr != SIN4(&proxy)->sin_addr.s_addr)
+		if (!sin_equal(faddr, &proxy))
 			return 1;
 
-		if (localn == SIN4(&proxy)->sin_addr.s_addr)
+		if (sin_equal(&localn_ss, &proxy))
 			return 1;
 	}
 
-	sin_setv4(htonl(localm), &ss);
-
-	ret = find_masq_entry(&ss, user, sizeof(user), os, sizeof(os));
+	ret = find_masq_entry(&localm_ss, user, sizeof(user), os, sizeof(os));
 
 	if (opt_enabled(FORWARD) && (ret != 0 || !opt_enabled(MASQ_OVERRIDE))) {
 		char ipbuf[MAX_IPLEN];
 
-		if (fwd_request(sock, lport, masq_lport, fport, masq_fport, &ss) == 0)
+		if (fwd_request(sock, lport, masq_lport, fport, masq_fport, &localm_ss) == 0)
 			return 0;
 
-		get_ip(&ss, ipbuf, sizeof(ipbuf));
+		get_ip(&localm_ss, ipbuf, sizeof(ipbuf));
 
 		debug("Forward to %s (%d %d) failed", ipbuf, masq_lport, fport);
 	}
@@ -732,7 +628,7 @@ static int masq_ct_line(char *line,
 
 		get_ip(faddr, ipbuf, sizeof(ipbuf));
 
-		o_log(NORMAL,
+		o_log(LOG_INFO,
 			"[%s] (Masqueraded) Successful lookup: %d , %d : %s",
 			ipbuf, lport, fport, user);
 
@@ -859,7 +755,7 @@ static uid_t lookup_tcp_diag(	struct sockaddr_storage *src_addr,
 				if (r->tcpdiag_inode == 0 && r->tcpdiag_uid == 0)
 					return MISSING_UID;
 
-				return (r->tcpdiag_uid);
+				return r->tcpdiag_uid;
 			}
 
 			return MISSING_UID;
@@ -876,7 +772,20 @@ static uid_t lookup_tcp_diag(	struct sockaddr_storage *src_addr,
 ** Just open a netlink socket here.
 */
 
+/*
+** Open the kernel memory device.
+** Return 0 on success, or -1 with errno set.
+**
+** No kmem access required; open a Netlink socket instead.
+*/
+
 int k_open(void) {
 	netlink_sock = socket(AF_NETLINK, SOCK_DGRAM, NETLINK_TCPDIAG);
-	return (0);
+
+	if (netlink_sock == -1) {
+		/* Not a fatal error, just log a debug message */
+		debug("Failed to open netlink socket: %s", strerror(errno));
+	}
+
+	return 0;
 }
